package main

import (
	"context"
	"net"
	"time"

	"gitlab.mashroom.com/mashroom/backend/schedule/internal/model"

	"github.com/go-redis/redis/v8"
	"github.com/zeebe-io/zeebe/clients/go/pkg/pb"
	"github.com/zeebe-io/zeebe/clients/go/pkg/zbc"
	"gitlab.mashroom.com/mashroom/backend/common/generic/addr"
	"gitlab.mashroom.com/mashroom/backend/common/generic/bpms"
	"gitlab.mashroom.com/mashroom/backend/common/generic/env"
	"gitlab.mashroom.com/mashroom/backend/common/generic/grpc"
	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	notification "gitlab.mashroom.com/mashroom/backend/notification/api"
	property "gitlab.mashroom.com/mashroom/backend/property/api"
	schedule "gitlab.mashroom.com/mashroom/backend/schedule/api"
	"gitlab.mashroom.com/mashroom/backend/schedule/bpmn"
	tenancyending "gitlab.mashroom.com/mashroom/backend/schedule/internal/handlers/tenancy/ending"
	tenancytermination "gitlab.mashroom.com/mashroom/backend/schedule/internal/handlers/tenancy/termination"
	redisrepository "gitlab.mashroom.com/mashroom/backend/schedule/internal/model/redis"
	"gitlab.mashroom.com/mashroom/backend/schedule/internal/service"
)

var (
	envRedisAddress  = env.ResolveString("REDIS_ADDR", "mashroom-schedule-redis:6379")
	envConsumerID    = env.ResolveString("CONSUMER_ID", "consumer-global")
	envDequeuePeriod = env.ResolveDuration("DEQUEUE_PERIOD", time.Second)

	propertyServiceURL = env.ResolveString("PROPERTY_SERVICE_URL", "mashroom-property:10000")
	propertyService    = property.NewPropertyServiceClient(grpc.Dial(propertyServiceURL, "Property"))

	notificationServiceURL = env.ResolveString("NOITIFCATION_SERVICE_URL", "mashroom-notification:10000")
	notificationService    = notification.NewNotificationServiceClient(grpc.Dial(notificationServiceURL, "Notification"))
)

// returns map of handlers to use in this application
func makeHandlers() (handlers map[string]service.Handler, err error) {
	handlers = make(map[string]service.Handler)

	handlers["tenancy_ending"], err = tenancyending.NewHandler(&tenancyending.Config{
		PropertyService:     propertyService,
		NotificationService: notificationService,
	})
	if err != nil {
		return nil, err
	}

	zeebeClient, err := makeZeebeClient()
	if err != nil {
		return nil, err
	}

	handlers["tenancy_termination"], err = tenancytermination.NewHandler(&tenancytermination.Config{
		PropertyService:     propertyService,
		NotificationService: notificationService,
		ZeebeClient:         zeebeClient,
	})
	if err != nil {
		return nil, err
	}

	return
}

func makeZeebeClient() (zbc.Client, error) {
	if bpms.URL == "" {
		return nil, nil
	}

	client, err := zbc.NewClient(&zbc.ClientConfig{
		UsePlaintextConnection: true,
		GatewayAddress:         bpms.URL,
	})
	if err != nil {
		return nil, err
	}

	bs := bpmn.MustAsset("tenancy-post-termination.bpmn")

	_, err = client.NewDeployWorkflowCommand().
		AddResource(bs, bpmn.ProcessTenancyPostTermination, pb.WorkflowRequestObject_BPMN).
		Send(context.Background())
	if err != nil {
		return nil, err
	}

	log.Infof("Workflow %v deployed", bpmn.ProcessTenancyPostTermination)

	return client, nil
}

func ensurePredefinedTask(ctx context.Context, repository model.Repository, task *model.Task) error {
	ex, err := repository.TaskGet(ctx, task.ID)
	if err != nil {
		return err
	}

	if ex == nil {
		ex = task
	}

	ex.Data = task.Data
	ex.Retry = task.Retry
	ex.Period = task.Period
	ex.Active = task.Active
	ex.Handler = task.Handler

	return repository.TaskSave(ctx, ex)
}

func predefineTasks(ctx context.Context, repository model.Repository) error {
	err := ensurePredefinedTask(ctx, repository, &model.Task{
		ID:      "predefined-notify-ending-tenancy",
		Handler: "tenancy_ending",
		Active:  true,
		Period:  time.Hour * 24,
		Data:    nil,
	})
	if err != nil {
		return err
	}

	return nil
}

func main() {
	redisClient := redis.NewClient(&redis.Options{
		Addr:     envRedisAddress,
		Password: "",
		DB:       0,
	})

	handlers, err := makeHandlers()
	if err != nil {
		panic(err)
	}

	repository, err := redisrepository.NewRepository(context.Background(), &redisrepository.Config{
		RedisClient: redisClient,
		Consumer:    envConsumerID,
		Handlers:    handlers,
	})
	if err != nil {
		panic(err)
	}

	err = predefineTasks(context.Background(), repository)
	if err != nil {
		panic(err)
	}

	svc, err := service.NewService(context.Background(), &service.Config{
		Handlers:      handlers,
		Repository:    repository,
		DequeuePeriod: envDequeuePeriod,
	})
	if err != nil {
		panic(err)
	}

	listener, err := net.Listen("tcp", addr.Grpc)
	if err != nil {
		panic(err)
	}

	srv, err := grpc.NewServer("schedule")
	if err != nil {
		panic(err)
	}

	schedule.RegisterScheduleServiceServer(srv, svc)

	err = srv.Serve(listener)
	if err != nil {
		panic(err)
	}
}
