package ending

import (
	"context"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	"gitlab.mashroom.com/mashroom/backend/common/generic/util"
	"gitlab.mashroom.com/mashroom/backend/common/generic/wrap"
	messaging "gitlab.mashroom.com/mashroom/backend/messaging/api"
	notification "gitlab.mashroom.com/mashroom/backend/notification/api"
	property "gitlab.mashroom.com/mashroom/backend/property/api"
	"gitlab.mashroom.com/mashroom/backend/schedule/internal/model"
)

func (handler *Handler) RetryPeriod() time.Duration {
	// There may be many tenancies and this task isn't supposed to run more than once a day, so
	return time.Hour
}

func (handler *Handler) ValidateData(data []byte) error {
	return nil
}

func (handler *Handler) fetchTenancy(ctx context.Context, tenancyID string) *property.Tenancy {
	tenancyResp, err := handler.config.PropertyService.GetTenancy(ctx, &property.GetTenancyRequest{By: &property.GetTenancyRequest_Id{Id: tenancyID}})
	if err != nil {
		log.WithError(err).Errorf("error get tenancy by ID: %v", tenancyID)
		return nil
	}

	if len(tenancyResp.GetTenancy()) == 0 {
		log.Error("get empty tenancy")
		return nil
	}

	tenancy := tenancyResp.Tenancy[0]
	if tenancy.GetProperty().GetValue() == "" {
		log.Errorf("in tenancy %v empty property field", tenancy.Id)
		return nil
	}

	if tenancy.GetOffer().GetValue() == "" {
		log.Errorf("in tenancy %v empty offer field", tenancy.Id)
		return nil
	}

	return tenancy
}

func (handler *Handler) Run(ctx context.Context, instance *model.Invocation) error {
	tenancyList, err := handler.config.PropertyService.ListTenancyNotNotified(ctx, &property.ListTenancyRequest{
		TenancyState:      &property.TenancyStateRequest{State: property.TenancyState_TENANCY_ACTIVE},
		IntentionNotified: &property.IntentionNotifiedRequest{IntentionNotified: false},
	})
	if err != nil {
		return err
	}

	if len(tenancyList.Tenancy) == 0 {
		return nil
	}

	log.Infof("TenancyEndingReminder get %v tenancy to processing", len(tenancyList.Tenancy))
	for _, tenancyID := range tenancyList.Tenancy {
		tenancy := handler.fetchTenancy(ctx, tenancyID)
		if tenancy == nil {
			continue
		}

		prop, err := handler.config.PropertyService.Get(ctx, &property.GetRequest{Id: tenancy.GetProperty().GetValue()})
		if err != nil {
			log.WithError(err).Errorf("error get property by ID: %v", tenancy.GetProperty().GetValue())
			continue
		}

		// notifying landlord
		landlord := prop.GetProperty().GetPeople().GetLandlord().GetValue()
		_, err = handler.config.NotificationService.SendEvent(ctx, &notification.SendEventRequest{
			User:     landlord,
			Event:    notification.EventType_EVENT_OFFER_TENANCY_NEAR_END_NOTIFY_LANDLORD,
			Channels: []notification.EventChannel{notification.EventChannel_EVENT_CHANNEL_PLATFORM, notification.EventChannel_EVENT_CHANNEL_EMAILV2},
			Embeds: []*notification.EventEmbed{
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
					EmbedValue: tenancy.GetProperty().GetValue(),
				},
			},
		})
		if err != nil {
			log.WithError(err).Errorf("error sent to landlord: %v from property: %v", landlord, tenancy.Property.Value)
			continue
		}

		// notifying tenants
		isTenantsNotified := false
		for _, t := range tenancy.GetTenants().GetValue() {
			_, err = handler.config.NotificationService.SendEvent(ctx, &notification.SendEventRequest{
				User:     t.GetUser().GetValue(),
				Event:    notification.EventType_EVENT_OFFER_TENANCY_NEAR_END_NOTIFY_TENANT,
				Channels: []notification.EventChannel{notification.EventChannel_EVENT_CHANNEL_PLATFORM, notification.EventChannel_EVENT_CHANNEL_EMAILV2},
				Embeds: []*notification.EventEmbed{
					{
						EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
						EmbedValue: tenancy.GetProperty().GetValue(),
					},
				},
			})
			if err != nil {
				log.WithError(err).Errorf("error sent to tenant: %v from property: %v", t.GetUser().GetValue(), tenancy.GetProperty().GetValue())
				break
			}
			isTenantsNotified = true
		}

		if !isTenantsNotified {
			continue
		}

		// set tenancy as notified
		_, err = handler.config.PropertyService.UpdateTenancy(ctx, &property.UpdateTenancyRequest{
			Tenancy: &property.TenancyRequest{
				Id:                wrap.StringValue(util.PtrString(tenancy.GetId())),
				IntentionNotified: &property.IntentionNotifiedRequest{IntentionNotified: true},
			},
		})
		if err != nil {
			log.WithError(err).Errorf("error set IntentionNotified: true for tenancy: %v", tenancy.GetId())
			continue
		}

	}

	return nil
}
