package termination

import (
	"context"
	"encoding/json"
	"fmt"
	"time"

	"github.com/zeebe-io/zeebe/clients/go/pkg/commands"
	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	messaging "gitlab.mashroom.com/mashroom/backend/messaging/api"
	notification "gitlab.mashroom.com/mashroom/backend/notification/api"
	property "gitlab.mashroom.com/mashroom/backend/property/api"
	"gitlab.mashroom.com/mashroom/backend/schedule/bpmn"

	"gitlab.mashroom.com/mashroom/backend/schedule/internal/model"
)

func (handler *Handler) RetryPeriod() time.Duration {
	// Notifications should be quickly sent and process started
	return time.Minute
}

func (handler *Handler) ValidateData(data []byte) error {
	var m map[string]string

	err := json.Unmarshal(data, &m)
	if err != nil {
		return err
	}

	tenancyID, ok := m["tenancy_id"]
	if !ok || tenancyID == "" {
		return fmt.Errorf("missing tenancy_id")
	}

	return nil
}

func (handler *Handler) Run(ctx context.Context, instance *model.Invocation) error {
	var data map[string]string

	err := json.Unmarshal(instance.Data, &data)
	if err != nil {
		return err
	}

	tenancyID, ok := data["tenancy_id"]
	if !ok {
		return fmt.Errorf("invalid tenancy id")
	}

	tenancyResp, err := handler.config.PropertyService.GetTenancy(ctx, &property.GetTenancyRequest{
		By: &property.GetTenancyRequest_Id{
			Id: tenancyID,
		},
	})
	if err != nil {
		return err
	}

	log.Infof("TenancyTerminationNotification get %v tenancy to processing", len(tenancyResp.Tenancy))

	if len(tenancyResp.GetTenancy()) == 0 {
		return fmt.Errorf("get empty tenancy")
	}

	tenancy := tenancyResp.Tenancy[0]
	if tenancy.GetProperty().GetValue() == "" {
		return fmt.Errorf("in tenancy %v empty property field", tenancy.Id)
	}

	if tenancy.GetOffer().GetValue() == "" {
		return fmt.Errorf("in tenancy %v empty offer field", tenancy.Id)
	}

	prop, err := handler.config.PropertyService.Get(ctx, &property.GetRequest{Id: tenancy.GetProperty().GetValue()})
	if err != nil {
		log.WithError(err).Errorf("error get property by ID: %v", tenancy.GetProperty().GetValue())
		return err
	}

	if handler.config.ZeebeClient != nil {
		var cmd commands.CreateInstanceCommandStep3

		cmd, err = handler.config.ZeebeClient.NewCreateInstanceCommand().
			BPMNProcessId(bpmn.ProcessTenancyPostTermination).
			LatestVersion().
			VariablesFromMap(map[string]interface{}{
				"tenancyID":  tenancyID,
				"propertyID": tenancy.GetProperty().GetValue(),
			})
		if err != nil {
			return err
		}

		_, err = cmd.Send(ctx)
		if err != nil {
			return err
		}
	}

	landlord := prop.GetProperty().GetPeople().GetLandlord().GetValue()
	_, err = handler.config.NotificationService.SendEvent(ctx, &notification.SendEventRequest{
		User:     landlord,
		Event:    notification.EventType_EVENT_OFFER_TENANCY_END_NOTIFY_LANDLORD,
		Channels: []notification.EventChannel{notification.EventChannel_EVENT_CHANNEL_PLATFORM, notification.EventChannel_EVENT_CHANNEL_EMAILV2},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: tenancy.GetProperty().GetValue(),
			},
		},
	})
	if err != nil {
		log.WithError(err).Errorf("error sent to landlord: %v from property: %v", landlord, tenancy.Property.Value)
		return err
	}

	for _, t := range tenancy.GetTenants().GetValue() {
		_, err = handler.config.NotificationService.SendEvent(ctx, &notification.SendEventRequest{
			User:     t.GetUser().GetValue(),
			Event:    notification.EventType_EVENT_OFFER_TENANCY_END_NOTIFY_TENANT,
			Channels: []notification.EventChannel{notification.EventChannel_EVENT_CHANNEL_PLATFORM, notification.EventChannel_EVENT_CHANNEL_EMAILV2},
			Embeds: []*notification.EventEmbed{
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
					EmbedValue: tenancy.GetProperty().GetValue(),
				},
			},
		})
		if err != nil {
			log.WithError(err).Errorf("error sent to tenant: %v from property: %v", t.GetUser().GetValue(), tenancy.GetProperty().GetValue())
			break
		}
	}

	return nil
}
