package redis

import (
	"context"
	"errors"
	"math"
	"strconv"
	"time"

	"github.com/go-redis/redis/v8"

	"github.com/bsm/redislock"
	"gitlab.mashroom.com/mashroom/backend/schedule/internal/model"
)

const streamgroup = "schedule"

// Implementation detail:
// Used keys:
// - seq:tasks                -- incremented sequence of task IDs
// - seq:invocations          -- incremented sequence of invocation IDs
// - map:tasks                -- currently registered tasks with task ID as a key
// - schedule:$HANDLER:set    -- unix-time-sorted set of invocations
// - schedule:$HANDLER:mutex  -- lock for moving invocation from :set to :stream
// - schedule:$HANDLER:stream -- stream of for consumer group to be dequeued
type repositoryImpl struct {
	config Config
	locker *redislock.Client
}

func (repo *repositoryImpl) TaskSave(ctx context.Context, task *model.Task) error {
	if task.ID == "" {
		id, err := repo.config.RedisClient.Incr(ctx, "seq:task").Result()
		if err != nil {
			return err
		}

		task.ID = strconv.FormatInt(id, 10)
	}

	return repo.config.RedisClient.HSet(ctx, "map:tasks", task.ID, task).Err()
}

func (repo *repositoryImpl) TaskList(ctx context.Context) (tasks []*model.Task, err error) {
	m, err := repo.config.RedisClient.HGetAll(ctx, "map:tasks").Result()
	if errors.Is(err, redis.Nil) {
		return nil, nil
	}

	if err != nil {
		return nil, err
	}

	for k, v := range m {
		var task model.Task

		err := task.UnmarshalBinary(([]byte)(v))
		if err != nil {
			return nil, err
		}

		task.ID = k

		tasks = append(tasks, &task)
	}

	return
}

func (repo *repositoryImpl) TaskGet(ctx context.Context, id string) (*model.Task, error) {
	v, err := repo.config.RedisClient.HGet(ctx, "map:tasks", id).Result()
	if errors.Is(err, redis.Nil) {
		return nil, nil
	}

	if err != nil {
		return nil, err
	}

	var task model.Task

	err = task.UnmarshalBinary(([]byte)(v))
	if err != nil {
		return nil, err
	}

	task.ID = id

	return &task, nil
}

func (repo *repositoryImpl) TaskCancel(ctx context.Context, id string) error {
	return repo.config.RedisClient.HDel(ctx, "map:tasks", id).Err()
}

func (repo *repositoryImpl) dequeueSchedule(ctx context.Context, setkey, streamkey, mutexkey, invocation string) error {
	lock, err := repo.locker.Obtain(ctx, mutexkey, time.Millisecond*500, nil)
	if err != nil {
		return err
	}

	defer func() {
		_ = lock.Release(ctx)
	}()

	err = repo.config.RedisClient.XAdd(ctx, &redis.XAddArgs{
		Stream: streamkey,
		Values: []string{"data", invocation},
	}).Err()
	if err != nil {
		return err
	}

	return repo.config.RedisClient.ZRem(ctx, setkey, invocation).Err()
}

func (repo *repositoryImpl) dequeuePending(
	ctx context.Context,
	streamkey, streamgroup string,
	retry time.Duration,
) (invocation *model.Invocation, err error) {
	xpend, err := repo.config.RedisClient.XPendingExt(ctx, &redis.XPendingExtArgs{
		Stream:   streamkey,
		Group:    streamgroup,
		Start:    "-",
		End:      "+",
		Count:    math.MaxInt64,
		Consumer: "",
	}).Result()
	if err != nil {
		return
	}

	var pend *redis.XPendingExt

	for _, x := range xpend {
		if x.Idle > retry {
			pend = &x
			break
		}
	}

	if pend == nil {
		return
	}

	var msgs []redis.XMessage

	msgs, err = repo.config.RedisClient.XClaim(ctx, &redis.XClaimArgs{
		Stream:   streamkey,
		Group:    streamgroup,
		Consumer: repo.config.Consumer,
		MinIdle:  retry,
		Messages: []string{pend.ID},
	}).Result()
	if err != nil {
		return
	}

	if len(msgs) < 1 {
		return
	}

	data := msgs[0].Values["data"]
	s, ok := data.(string)
	if !ok {
		return
	}

	invocation = &model.Invocation{}

	err = invocation.UnmarshalBinary(([]byte)(s))
	if err != nil {
		return nil, err
	}

	invocation.AckID = msgs[0].ID

	if invocation.Retry >= int(pend.RetryCount) {
		return
	}

	err = repo.config.RedisClient.XDel(ctx, streamkey, msgs[0].ID).Err()
	if err != nil {
		return nil, err
	}

	invocation = nil

	return
}

func (repo *repositoryImpl) dequeueNext(
	ctx context.Context,
	streamkey, streamgroup string,
) (invocation *model.Invocation, err error) {
	xs, err := repo.config.RedisClient.XReadGroup(ctx, &redis.XReadGroupArgs{
		Group:    streamgroup,
		Consumer: repo.config.Consumer,
		Streams:  []string{streamkey, ">"},
		Count:    1,
		Block:    -1,
		NoAck:    false,
	}).Result()
	if errors.Is(err, redis.Nil) {
		return nil, nil
	}

	if err != nil {
		return
	}

	if len(xs) < 1 {
		return
	}

	if len(xs[0].Messages) < 1 {
		return
	}

	data := xs[0].Messages[0].Values["data"]
	s, ok := data.(string)
	if !ok {
		return
	}

	invocation = &model.Invocation{}

	err = invocation.UnmarshalBinary(([]byte)(s))
	if err != nil {
		return nil, err
	}

	invocation.AckID = xs[0].Messages[0].ID

	return
}

func (repo *repositoryImpl) InvocationDequeue(ctx context.Context, handler string, retry time.Duration) (invocation *model.Invocation, err error) {
	now := time.Now().Unix()

	keybase := "schedule:" + handler
	setkey := keybase + ":set"
	streamkey := keybase + ":stream"

	vs, err := repo.config.RedisClient.ZRangeByScore(ctx, setkey, &redis.ZRangeBy{
		Min:    "0",
		Max:    strconv.FormatInt(now, 10),
		Offset: 0,
		Count:  1,
	}).Result()
	if err != nil {
		return
	}

	if len(vs) == 1 {
		mutexkey := keybase + ":mutex"

		err = repo.dequeueSchedule(ctx, setkey, streamkey, mutexkey, vs[0])
		if err != nil {
			return
		}
	}

	invocation, err = repo.dequeuePending(ctx, streamkey, streamgroup, retry)
	if err != nil || invocation != nil {
		return
	}

	invocation, err = repo.dequeueNext(ctx, streamkey, streamgroup)

	return
}

func (repo *repositoryImpl) InvocationAck(ctx context.Context, handler string, ackid string) error {
	return repo.config.RedisClient.XAck(ctx, "schedule:"+handler+":stream", streamgroup, ackid).Err()
}

func (repo *repositoryImpl) InvocationCancel(ctx context.Context, handler string, id string) error {
	setkey := "schedule:" + handler + ":set"

	iter, _, err := repo.config.RedisClient.ZScan(ctx, setkey, 0, "*\"id\":\""+id+"\"*", 1).Result()
	if err != nil {
		return err
	}

	if len(iter) == 0 {
		return nil
	}

	return repo.config.RedisClient.ZRem(ctx, setkey, iter[0]).Err()
}

func (repo *repositoryImpl) InvocationGet(ctx context.Context, handler string, id string) (invocation *model.Invocation, err error) {
	setkey := "schedule:" + handler + ":set"

	iter, _, err := repo.config.RedisClient.ZScan(ctx, setkey, 0, "*\"id\":\""+id+"\"*", 1).Result()
	if err != nil {
		return
	}

	if len(iter) == 0 {
		return
	}

	invocation = &model.Invocation{}

	err = invocation.UnmarshalBinary(([]byte)(iter[0]))
	if err != nil {
		return nil, err
	}

	return
}

func (repo *repositoryImpl) InvocationEnqueue(ctx context.Context, handler string, invocation *model.Invocation) error {
	setkey := "schedule:" + handler + ":set"

	id, err := repo.config.RedisClient.Incr(ctx, "seq:invocation").Result()
	if err != nil {
		return err
	}

	invocation.ID = strconv.FormatInt(id, 10)

	return repo.config.RedisClient.ZAddNX(ctx, setkey, &redis.Z{
		Score:  float64(invocation.Instant.Unix()),
		Member: invocation,
	}).Err()
}
