package service

import (
	"context"
	"fmt"
	"time"

	schedule "gitlab.mashroom.com/mashroom/backend/schedule/api"
	"gitlab.mashroom.com/mashroom/backend/schedule/internal/model"
)

// Handler provides implementation for schedulable activity
type Handler interface {
	// Run is being invoked as execution of a scheduled activity
	Run(ctx context.Context, instance *model.Invocation) error
	// RetryPeriod indicated time this handler is expected to complete it's task
	// used for determining period after which task taken by a worker of this service would be reassigned to another
	RetryPeriod() time.Duration
	// Validate data ensures that data for this handler is correct
	ValidateData(data []byte) error
}

// Config for service implementation
type Config struct {
	// Handlers map from handler code to it's instance
	Handlers map[string]Handler
	// Persistent repository
	Repository model.Repository
	// Interval between dequeueing for each of handlers
	DequeuePeriod time.Duration
}

// Service implementation
type Service interface {
	schedule.ScheduleServiceServer
}

func validateConfig(config *Config) error {
	if config == nil {
		return fmt.Errorf("invalid service config")
	}

	if len(config.Handlers) == 0 {
		return fmt.Errorf("no handlers in service config")
	}

	if config.Repository == nil {
		return fmt.Errorf("no repository in service config")
	}

	if config.DequeuePeriod == 0 {
		return fmt.Errorf("zero dequeue period in service config")
	}

	return nil
}

// takes all active tasks and ensures that invocation is scheduled according to task's period
func (srv *serviceImpl) rearmActiveTasks(ctx context.Context) error {
	tasks, err := srv.config.Repository.TaskList(ctx)
	if err != nil {
		return err
	}

	for _, t := range tasks {
		if !t.Active {
			continue
		}

		var inv *model.Invocation

		if t.Armed != "" {
			inv, err = srv.config.Repository.InvocationGet(ctx, t.Handler, t.Armed)
			if err != nil {
				return err
			}
		}

		if inv == nil {
			inv = &model.Invocation{
				Instant: time.Now().Add(t.Period),
				Data:    t.Data,
				Retry:   t.Retry,
				Task:    t.ID,
			}

			err = srv.config.Repository.InvocationEnqueue(ctx, t.Handler, inv)
			if err != nil {
				return err
			}

			t.Armed = inv.ID

			err = srv.config.Repository.TaskSave(ctx, t)
			if err != nil {
				return err
			}
		}
	}

	return nil
}

// NewService returns new service instance, initializing pending tasks and starting background dequeuers
func NewService(ctx context.Context, config *Config) (Service, error) {
	err := validateConfig(config)
	if err != nil {
		return nil, err
	}

	srv := &serviceImpl{
		config: *config,
	}

	err = srv.rearmActiveTasks(ctx)
	if err != nil {
		return nil, err
	}

	for k, h := range config.Handlers {
		go srv.handlerDequeue(ctx, k, h)
	}

	return srv, nil
}
