package service

import (
	"context"
	"time"

	"gitlab.mashroom.com/mashroom/backend/schedule/internal/model"

	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
)

// attempt rearming recurrent task
func (srv *serviceImpl) attemptRearmTask(ctx context.Context, key string, inv *model.Invocation) {
	t, err := srv.config.Repository.TaskGet(ctx, inv.Task)
	if err != nil {
		log.WithError(err).WithField("invocation", inv).Errorf("fetching task for key %s", key)
		return
	}

	if t.Armed != inv.ID {
		return
	}

	nextinv := &model.Invocation{
		Instant: time.Now().Add(t.Period),
		Data:    t.Data,
		Retry:   t.Retry,
		Task:    t.ID,
	}

	err = srv.config.Repository.InvocationEnqueue(ctx, key, nextinv)
	if err != nil {
		log.WithError(err).WithField("invocation", inv).Errorf("rearming for key %s", key)
	}

	t.Armed = nextinv.ID

	err = srv.config.Repository.TaskSave(ctx, t)
	if err != nil {
		log.WithError(err).WithField("invocation", inv).Errorf("updating task for key %s", key)
	}
}

// polls single activity for execution
func (srv *serviceImpl) handlerDequeueNext(ctx context.Context, key string, handler Handler) {
	inv, err := srv.config.Repository.InvocationDequeue(context.Background(), key, handler.RetryPeriod())
	if err != nil {
		log.WithError(err).Errorf("dequeueing for key %s", key)
		return
	}

	if inv == nil {
		return
	}

	ctx, cancel := context.WithTimeout(ctx, handler.RetryPeriod())

	defer cancel()

	if inv.Task != "" {
		srv.attemptRearmTask(ctx, key, inv)
	}

	err = handler.Run(ctx, inv)
	if err != nil {
		log.WithError(err).WithField("invocation", inv).Errorf("invoking for key %s", key)
	}

	err = srv.config.Repository.InvocationAck(ctx, key, inv.AckID)
	if err != nil {
		log.WithError(err).WithField("invocation", inv).Errorf("acking for key %s", key)
	}
}

// polls activities for execution until parent context is cancelled
func (srv *serviceImpl) handlerDequeue(ctx context.Context, key string, handler Handler) {
	ticker := time.NewTicker(srv.config.DequeuePeriod)

	for {
		select {
		case <-ctx.Done():
			return
		case <-ticker.C:
			srv.handlerDequeueNext(ctx, key, handler)
		}
	}
}
