package service

import (
	"context"
	"encoding/json"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/api/grpcerr"
	schedule "gitlab.mashroom.com/mashroom/backend/schedule/api"
	"gitlab.mashroom.com/mashroom/backend/schedule/internal/model"
)

var errUnknownHandler = grpcerr.New(grpcerr.InvalidData, "unknown handler")

type serviceImpl struct {
	config Config
}

func (srv *serviceImpl) InvocationSchedule(ctx context.Context, in *schedule.InvocationScheduleRequest) (*schedule.InvocationScheduleResponse, error) {
	h, ok := srv.config.Handlers[in.GetHandler()]
	if !ok {
		return nil, errUnknownHandler
	}

	bs, err := json.Marshal(in.GetData())
	if err != nil {
		return nil, err
	}

	err = h.ValidateData(bs)
	if err != nil {
		return nil, err
	}

	inv := &model.Invocation{
		Instant: in.GetInstant().AsTime(),
		Retry:   int(in.GetRetry()),
		Data:    bs,
	}

	err = srv.config.Repository.InvocationEnqueue(ctx, in.GetHandler(), inv)
	if err != nil {
		return nil, err
	}

	return &schedule.InvocationScheduleResponse{
		InvocationId: inv.ID,
	}, nil
}

func (srv *serviceImpl) InvocationCancel(ctx context.Context, in *schedule.InvocationCancelRequest) (*schedule.InvocationCancelResponse, error) {
	err := srv.config.Repository.InvocationCancel(ctx, in.GetHandler(), in.GetInvocationId())
	if err != nil {
		return nil, err
	}

	return &schedule.InvocationCancelResponse{}, nil
}

func (srv *serviceImpl) TaskList(ctx context.Context, in *schedule.TaskListRequest) (*schedule.TaskListResponse, error) {
	tasks, err := srv.config.Repository.TaskList(ctx)
	if err != nil {
		return nil, err
	}

	scheduleTasks, err := renderTasks(tasks)
	if err != nil {
		return nil, err
	}

	return &schedule.TaskListResponse{
		Tasks: scheduleTasks,
	}, nil
}

func (srv *serviceImpl) TaskGet(ctx context.Context, in *schedule.TaskGetRequest) (*schedule.TaskGetResponse, error) {
	task, err := srv.config.Repository.TaskGet(ctx, in.GetId())
	if err != nil {
		return nil, err
	}

	scheduleTask, err := renderTask(task)
	if err != nil {
		return nil, err
	}

	return &schedule.TaskGetResponse{
		Task: scheduleTask,
	}, nil
}

func (srv *serviceImpl) TaskSave(ctx context.Context, in *schedule.TaskSaveRequest) (*schedule.TaskSaveResponse, error) {
	data, err := json.Marshal(in.GetData())
	if err != nil {
		return nil, err
	}

	h, ok := srv.config.Handlers[in.GetHandler()]
	if !ok {
		return nil, errUnknownHandler
	}

	err = h.ValidateData(data)
	if err != nil {
		return nil, err
	}

	var task *model.Task

	if in.GetId() != "" {
		task, err = srv.config.Repository.TaskGet(ctx, in.GetId())
		if err != nil {
			return nil, err
		}
	} else {
		task = &model.Task{}
	}

	task.Handler = in.GetHandler()
	task.Active = in.GetActive()
	task.Period = in.GetPeriod().AsDuration()
	task.Data = data
	task.Retry = int(in.GetRetry())

	if task.Active && task.Armed == "" {
		inv := &model.Invocation{
			Instant: time.Now().Add(task.Period),
			Data:    task.Data,
			Retry:   task.Retry,
			Task:    task.ID,
		}

		err = srv.config.Repository.InvocationEnqueue(ctx, in.GetHandler(), inv)
		if err != nil {
			return nil, err
		}

		task.Armed = inv.ID
	} else if !task.Active && task.Armed != "" {
		err = srv.config.Repository.InvocationCancel(ctx, task.Handler, task.Armed)
		if err != nil {
			return nil, err
		}

		task.Armed = ""
	}

	err = srv.config.Repository.TaskSave(ctx, task)
	if err != nil {
		return nil, err
	}

	scheduleTask, err := renderTask(task)
	if err != nil {
		return nil, err
	}

	return &schedule.TaskSaveResponse{
		Task: scheduleTask,
	}, nil
}

func (srv *serviceImpl) TaskCancel(ctx context.Context, in *schedule.TaskCancelRequest) (*schedule.TaskCancelResponse, error) {
	err := srv.config.Repository.TaskCancel(ctx, in.GetId())
	if err != nil {
		return nil, err
	}

	return &schedule.TaskCancelResponse{}, nil
}
